package rabbit.webserver;

import java.io.File;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.URL;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Date;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import rabbit.http.HttpDateParser;
import rabbit.http.HttpHeader;
import rabbit.httpio.Acceptor;
import rabbit.httpio.AcceptorListener;
import rabbit.httpio.FileResourceSource;
import rabbit.httpio.HttpHeaderListener;
import rabbit.httpio.HttpHeaderReader;
import rabbit.httpio.HttpHeaderSender;
import rabbit.httpio.HttpHeaderSentListener;
import rabbit.httpio.ResourceSource;
import rabbit.httpio.SelectorRunner;
import rabbit.httpio.TransferHandler;
import rabbit.httpio.TransferListener;
import rabbit.io.BufferHandle;
import rabbit.io.BufferHandler;
import rabbit.io.CacheBufferHandle;
import rabbit.io.CachingBufferHandler;
import rabbit.util.Logger;
import rabbit.util.SimpleLogger;
import rabbit.util.SimpleTrafficLogger;
import rabbit.util.TrafficLogger;

/** A simple web server that serves static resources.
 *
 * @author <a href="mailto:robo@khelekore.org">Robert Olofsson</a>
 */
public class SimpleWebServer {
    private File dir;
    private int port;
    private final Logger logger = new SimpleLogger ();
    private final SelectorRunner selectorRunner;
    private final TrafficLogger trafficLogger = new SimpleTrafficLogger ();
    private final BufferHandler bufferHandler = new CachingBufferHandler ();
    
    public static void main (String[] args) {
	if (args.length != 2) {
	    usage ();
	    return;
	}
	try {
	    int port = Integer.parseInt (args[0]);
	    SimpleWebServer sws = new SimpleWebServer (port, args[1]);
	    sws.start ();
	} catch (IOException e) {
	    e.printStackTrace ();
	}
    }

    private static void usage () {
	System.err.println ("java " + SimpleWebServer.class.getName () + 
			    " <port> <dir>");
    }

    public SimpleWebServer (int port, String path) throws IOException {
	this.port = port;
	dir = new File (path);
	if (!(dir.exists () && dir.isDirectory ()))
	    throw new IOException (dir + " is not an existing directory");
	dir = dir.getCanonicalFile ();
	ExecutorService es = Executors.newCachedThreadPool ();
	selectorRunner = new SelectorRunner (es, logger);
    }

    public void start ()  {
	selectorRunner.start ();
	selectorRunner.runMainTask (new Runnable () {
		public void run () {
		    setupServerSocket ();
		}
	    });	
    }

    private void setupServerSocket () {
	try {
	    ServerSocketChannel ssc = ServerSocketChannel.open ();
	    ssc.configureBlocking (false);
	    ssc.socket ().bind (new InetSocketAddress (port));
	    AcceptorListener acceptListener = new AcceptListener ();
	    Acceptor acceptor = 
		new Acceptor (ssc, selectorRunner, logger, acceptListener);
	    acceptor.register ();
	} catch (IOException e) {
	    throw new RuntimeException ("Failed to setup server socket", e);
	}
    }

    private class AcceptListener implements AcceptorListener {
	public void connectionAccepted (SocketChannel sc, Selector selector) 
	    throws IOException {
	    new Connection (sc).readRequest ();
	}
    }

    private class Connection {
	private SocketChannel sc;
	private BufferHandle clientBufferHandle;
	private boolean timeToClose = false;
	private ResourceSource resourceSource = null;

	public Connection (SocketChannel sc) {
	    this.sc = sc;
	}

	public void readRequest () throws IOException {
	    if (clientBufferHandle == null)
		clientBufferHandle = new CacheBufferHandle (bufferHandler);
	    HttpHeaderListener requestListener = new RequestListener ();
	    new HttpHeaderReader (sc, clientBufferHandle, 
				  selectorRunner.getSelector (),
				  logger, trafficLogger, true,
				  true, requestListener);
	}

	private void shutdown () {
	    try {
		sc.close ();
	    } catch (IOException e) {
		logger.logWarn ("Failed to close down connection: " + e);
	    }
   	}
	
	private void handleRequest (HttpHeader header) {
	    String method = header.getMethod ();
	    if ("GET".equals (method) || "HEAD".equals (method)) {
		String path = header.getRequestURI ();
		if (path == null || "".equals (path)) {
		    badRequest ();
		    return;
		} 
		try {
		    if (!path.startsWith ("/")) {
			URL u = new URL (path);
			path = u.getFile ();
		    }
		    if (path.endsWith ("/"))
			path += "index.html";
		    path = path.substring (1);
		    File f = new File (dir, path);
		    f = f.getCanonicalFile ();
		    if (isSafe (f) && f.exists () && f.isFile ()) {
			HttpHeader resp = getHeader ("HTTP/1.1 200 Ok");
			resp.setHeader ("Content-Length", 
					Long.toString (f.length ()));
			if ("HTTP/1.0".equals (header.getHTTPVersion ()))
			    resp.setHeader ("Connection", "Keep-Alive");
			if ("GET".equals (method))
			    resourceSource = 
				new FileResourceSource (f, selectorRunner, 
							bufferHandler);
			sendResponse (resp);
		    } else {
			notFound ();
		    }
		} catch (IOException e) {
		    internalError ();
		}
	    } else {
		methodNotAllowed ();
	    }
	}

	private boolean isSafe (File f) {
	    return f.getAbsolutePath ().startsWith (dir.getAbsolutePath ());
	}

	private void notFound () {
	    sendResponse (getHeader ("HTTP/1.1 404 Not Found"));
	}

	private void badRequest () {
	    sendBadResponse (getHeader ("HTTP/1.1 400 Bad Request"));
	}
	
	private void methodNotAllowed () {
	    sendBadResponse (getHeader ("HTTP/1.1 405 Method Not Allowed"));
	}

	private void internalError () {
	    sendBadResponse (getHeader ("HTTP/1.1 500 Internal Error"));
	}

	private void notImplemented () {
	    sendBadResponse (getHeader ("HTTP/1.1 501 Not Implemented"));
	}

	private void sendBadResponse (HttpHeader response) {
	    timeToClose = true;
	    sendResponse (response);
	}

	private void sendResponse (HttpHeader response) {
	    try {
		ResponseSentListener sentListener = 
		    new ResponseSentListener ();
		new HttpHeaderSender (sc, selectorRunner.getSelector (), 
				      logger, trafficLogger, response, 
				      false, sentListener);
	    } catch (IOException e) {
		shutdown ();
	    }
	}

	private void sendResource () {
	    TransferListener transferDoneListener = 
		new TransferDoneListener ();
	    TransferHandler th = 
		new TransferHandler (selectorRunner, resourceSource, 
				     selectorRunner.getSelector (), sc, 
				     trafficLogger, trafficLogger, 
				     transferDoneListener, logger);
	    th.transfer ();
	}

	public HttpHeader getHeader (String statusLine) {
	    HttpHeader ret = new HttpHeader ();
	    ret.setStatusLine (statusLine);
	    ret.setHeader ("Server", SimpleWebServer.this.getClass ().getName ());
	    ret.setHeader ("Content-type", "text/html");
	    ret.setHeader ("Date", HttpDateParser.getDateString (new Date ()));
	    return ret;    
	}

	private void closeOrContinue () {
	    if (timeToClose) {
		shutdown ();
	    } else {
		try {
		    readRequest ();
		} catch (IOException e) {
		    shutdown ();
		}
	    }
	}

	private class AsyncBaseListener {
	    public void timeout () {
		shutdown ();
	    }
	    
	    public void failed (Exception e) {
		shutdown ();
	    }
	}
	
	private class RequestListener extends AsyncBaseListener 
	    implements HttpHeaderListener {
	    public void httpHeaderRead (HttpHeader header, BufferHandle bh, 
					boolean keepalive, boolean isChunked, 
					long dataSize) {
		bh.possiblyFlush ();
		if (isChunked || dataSize > 0)
		    notImplemented ();
		if (!keepalive)
		    timeToClose = true;
		handleRequest (header);
	    }

	    public void closed () {
		shutdown ();
	    }
	}

	private class ResponseSentListener extends AsyncBaseListener 
	    implements HttpHeaderSentListener {
	    public void httpHeaderSent () {
		if (resourceSource != null)
		    sendResource ();
		else  
		    closeOrContinue ();
	    }
	}

	private class TransferDoneListener extends AsyncBaseListener 
	    implements TransferListener {
	    public void transferOk () {
		resourceSource.release ();
		closeOrContinue ();	
	    }
	}
    }
}
